#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/hdreg.h>	/* HDIO_GETGEO */
#include <linux/kdev_t.h>
#include <linux/vmalloc.h>
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/buffer_head.h>	/* invalidate_bdev */

#define inb(x) readb(x)
#define outb(x,y) writeb(x,y)
#define inw(x) readw(x)
#define outw(x,y) writew(x,y)
#define inl(x) readl(x)
#define outl(x,y) writel(x,y)
#define SDDAT           1
extern void __udelay(unsigned int);

#define SD_DEFNAME		"tssdcard"
#define SD_DEFIOSIZE		4
#define SD_BLKSIZE		1024	//nsectors?
#include "tssdcore2.c"

#define TS7XXX_SDCARD1		0xE8000100
#define TS7800_FPGASDDAT	0x80000104
#define TS7800_CPUDOORBELL	0xF1020400
#define TS7800_FPGADOORBELL	0xE8000200
#define TS7800_FPGAMSIMASK	0xE8000204
static unsigned char *name = SD_DEFNAME;
static unsigned int io = TS7XXX_SDCARD1;
	
static struct block_device_operations tssdcard_ops;
static void tssdcard_delay(void *arg, unsigned int us)
{
	__udelay(us);
}

/* Minor/Major info and partition management */
static int tssdcard_major = 0;		/* dynamically allocate one */
#define KERNEL_SECTOR_SIZE	512
#define SD_HARDSEC		512
#define TSSDCARD_MINORS		SD_NUMPART
#define SD_NUMPART		16
#define MINOR_SHIFT		4

/* Internal representation of our device */
struct tssdcard_dev {
	struct sdcore tssdcore;
	char *devname;
	int size;		/* Device size in sectors */
	struct gendisk *gd;	/* The gendisk structure */
	short users;
	int parksect;
	struct request_queue *queue;	/* The device request queue */
	spinlock_t lock;
};
static struct tssdcard_dev *devices = NULL;


/* Handle an I/O request.*/
static inline void tssdcard_transfer(struct tssdcard_dev *dev,
				     unsigned long sector,
				     unsigned long nsect, char *buffer,
				     int rw)
{
	int ret;
	if ((sector + nsect) > (dev->size * 512 / KERNEL_SECTOR_SIZE)) {
		printk(KERN_NOTICE "Beyond-end write (%ld %ld)\n", sector,
			nsect);
		return;
	}

	switch (rw) {
	case WRITE:
		ret = sdwrite(&dev->tssdcore, sector, buffer, nsect);
		if (ret && !dev->tssdcore.sd_wprot) 
			printk("SD: write FAILED!\n n");
		break;
	case READ:
	case READA:
		ret = sdread(&dev->tssdcore, sector, buffer, nsect);
		if (ret && !dev->tssdcore.sd_wprot) 
			printk("SD: read FAILED!\n n");
	
		dev->parksect = sector + nsect - 1;

		//sdread will park at sector+nsect+1, if we read the last sector
		//this means we are parked on an invalid sector so reset the card.
		if ((sector + nsect) >= (dev->size * 512 / KERNEL_SECTOR_SIZE)) {
			printk
			    ("SD: sdread left card parked at invalid sector, reseting card\n");
			sdreset(&dev->tssdcore);
		}
		break;
	}
}

/* Our request function */
static void tssdcard_request(request_queue_t * q)
{
	struct request *req;
	while ((req = elv_next_request(q)) != NULL) {
		struct tssdcard_dev *dev = q->queuedata;
		if (!blk_fs_request(req)) {
			printk(KERN_NOTICE "Skip non-fs request\n");
			end_request(req, 0);
			continue;
		}
		tssdcard_transfer(dev, req->sector,
				   req->current_nr_sectors, req->buffer,
				   rq_data_dir(req));
		end_request(req, 1);
	}
}

/* Open and close. */
static int tssdcard_open(struct inode *inode, struct file *filp)
{
	struct tssdcard_dev *dev = inode->i_bdev->bd_disk->private_data;
	filp->private_data = dev;

	spin_lock(&dev->lock);
	//if(!dev->users) check_disk_change(inode->i_bdev);
	dev->users++;
	spin_unlock(&dev->lock);

	return 0;
}
static int tssdcard_release(struct inode *inode, struct file *filp)
{
	struct tssdcard_dev *dev = inode->i_bdev->bd_disk->private_data;

	spin_lock(&dev->lock);
	dev->users--;
	spin_unlock(&dev->lock);

	return 0;
}

static void tssdcard_sdcommit(unsigned long ldev)
{
	struct tssdcard_dev *dev = (struct tssdcard_dev *) ldev;
	char buf[512];
	sdread(&dev->tssdcore, dev->parksect, buf, 1);
	return;
}

static int tssdcard_shutdownhook(struct notifier_block *NotifierBlock,
    unsigned long Event, void *Buffer)
{
	int i;

	if (!(Event == SYS_RESTART || Event == SYS_HALT || Event == SYS_POWER_OFF))
		return NOTIFY_DONE;

	for (i = 0; i< 2 ; i++)
		tssdcard_sdcommit((unsigned long)(devices+i));

	return NOTIFY_OK;
}

static struct notifier_block shutdownhook_notifier = { tssdcard_shutdownhook, NULL, 0 };

/*
 * The ioctl() implementation
 */
int tssdcard_ioctl(struct inode *inode, struct file *filp,
		   unsigned int cmd, unsigned long arg)
{
	struct hd_geometry geo;
	struct tssdcard_dev *dev = filp->private_data;
	switch (cmd) {
	case HDIO_GETGEO:

		/*
		 * get geometry: we have to fake one...  trim the size to a
		 * multiple of 512 (256k): tell we have 32 sectors, 16 heads,
		 * whatever cylinders.
		 */
		geo.cylinders = dev->size / (16 * 32);
		geo.heads = 16;
		geo.sectors = 32;
		geo.start = 0;
		if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
			return -EFAULT;
		return 0;
	}
	return -ENOTTY;	/* unknown command */
}


/*
 * The device operations structure.
 */
static struct block_device_operations tssdcard_ops = { 
	.owner = THIS_MODULE, 
	.open = tssdcard_open, 
	.release = tssdcard_release, 
	.ioctl = tssdcard_ioctl
};


/*
 * Set up our internal device.
 */
static void setup_device(struct tssdcard_dev *dev, int which)
{
	static unsigned long addr = 0;
	unsigned int scratchreg0;
	unsigned long conreg;
	printk
	    ("%s: Technologic Systems SD card controller, address 0x%x\n",
	     name, io);
	memset(dev, 0, sizeof(struct tssdcard_dev));

	//IO Remapping (use the same virtual address for all LUNs)
	if (!addr)
		addr = (unsigned long) ioremap(io, SD_DEFIOSIZE);

	dev->tssdcore.sd_regstart = addr;
	dev->tssdcore.sd_lun = which;

	//verify previous init
	conreg = (unsigned long) 0;
	scratchreg0 = 0;
	if (scratchreg0) {
		dev->tssdcore.sdboot_token = scratchreg0;
		outl(0, 0);
	}
	
	dev->tssdcore.os_arg = dev;
	dev->tssdcore.os_delay = tssdcard_delay;
	dev->tssdcore.os_dmastream = NULL;
	dev->tssdcore.os_dmaprep = NULL;
	dev->tssdcore.sd_writeparking = 1;
	dev->parksect = 0;
	dev->users = 0;
	dev->size = sdreset(&dev->tssdcore);

	if (dev->size == 0) {
		printk("%s: no card found.\n", name);
		return;
	}

	spin_lock_init(&dev->lock);

	dev->queue = blk_init_queue(tssdcard_request, &dev->lock);
	if (dev->queue == NULL) return;

	blk_queue_hardsect_size(dev->queue, 512);
	dev->queue->queuedata = dev;

	dev->gd = alloc_disk(TSSDCARD_MINORS);
	if (!dev->gd) {
		printk(KERN_NOTICE "alloc_disk failure\n");
		return;
	}

	dev->devname = kmalloc(32, GFP_KERNEL);
	strcpy(dev->devname, name);
	snprintf(dev->gd->disk_name, 32, strcat(dev->devname, "%c"),
		  which + 'a');
	strcpy(dev->devname, dev->gd->disk_name);
	printk("%s: card /dev/%s has %d sectors (LUN %d)\n", name,
		dev->devname, dev->size, which);


	dev->gd->major = tssdcard_major;
	dev->gd->first_minor = which * TSSDCARD_MINORS;
	dev->gd->flags = GENHD_FL_REMOVABLE;
	dev->gd->fops = &tssdcard_ops;
	dev->gd->queue = dev->queue;
	dev->gd->private_data = dev;
	set_capacity(dev->gd, dev->size * 512 / KERNEL_SECTOR_SIZE);
	add_disk(dev->gd);

	return;
}
static int __init tssdcard_init(void)
{
	int i;
	tssdcard_major = register_blkdev(tssdcard_major, name);
	if (tssdcard_major <= 0) {
		printk(KERN_WARNING
			"tssdcard: unable to get major number\n");
		return -EBUSY;
	}
	devices =
	    kmalloc(2 * sizeof(struct tssdcard_dev), GFP_KERNEL);
	if (devices == NULL)
		goto out_unregister;

	for (i = 0; i < 2; i++)
		setup_device(devices + i, i);

	return 0;

	out_unregister:unregister_blkdev(tssdcard_major, name);
	return -ENOMEM;
}
static void tssdcard_exit(void)
{
	struct tssdcard_dev *dev;
	int i;
	for (i = 0; i < 2; i++) {
		dev = devices + i;
		if (dev == NULL)
			continue;
		if (dev->gd) {
			del_gendisk(dev->gd);
			put_disk(dev->gd);
		}
		if (dev->queue) {
			blk_cleanup_queue(dev->queue);
		}
		if (i == 0)
			iounmap((unsigned long *) dev->tssdcore.
				sd_regstart);
	} 
	unregister_blkdev(tssdcard_major, name);
	kfree(devices);
} module_init(tssdcard_init);

module_exit(tssdcard_exit);
MODULE_LICENSE("GPL");
