/*
 * Copyright (c) 2006, Technologic Systems
 * All rights reserved.
 */

/* Register offset definitions.  TS-SDCORE is 4 8-bit regs total. */
#define SDCMD		0
#define SDDAT		1
#define SDSTATE		2
#define SDCTRL		3

struct sdcore {
	/* virtual address of SD block register start, to be filled in
	 * by client code before calling any sdcore functions.
	 */
	unsigned int sd_regstart;

	/* public bits for sd_state bitfield, can be read from client code.
	 * Do not write!  Other bits are used internally.
	 */
	#define SDDAT_RX	(1<<0)
	#define SDDAT_TX	(1<<1)
	#define SDCMD_RX	(1<<2)
	#define SDCMD_TX	(1<<3)
	unsigned int sd_state;

	/* Erase hint for subsequent sdwrite() call, used to optimize
	 * write throughput on multi-sector writes by pre-erasing this
	 * many sectors.
	 */
	unsigned int sd_erasehint;

	/* Following this comment are 3 function pointer declarations to
	 * OS helper functions.  The 'os_arg' member is passed as the
	 * first argument to the helpers and should be set by
	 * client code before issueing sdreset()
	 *
	 * os_dmastream(os_arg, buf, buflen)
	 * This function should look at sd_state and set up and run an
	 * appropriate DMA transfer.  If buf is NULL, callee doesn't care
	 * about the actual data sent/received and helper function
	 * can do whatever it wants.  Should return 0 when DMA transfer was
	 * run and completed successfully.  If this function pointer is
	 * NULL, PIO methods of transfer will be used instead of DMA.
	 *
	 * os_dmaprep(os_arg, buf, buflen)
	 * This function is used to prepare an area of memory for a possible
	 * DMA transfer.  This function is called once per distinct buffer
	 * passed in.  After this function is called, os_dmastream() may be
	 * called one or more times (for sequential addresses) on subregions
	 * of the address range passed here.  Should write-back or invalidate
	 * L1 cache lines and possibly look up physical addresses for buf
	 * passed in if I/O buffers.  If 'os_dmaprep' is set to NULL, function
	 * call will not happen. (though os_dmastream() calls may still)
	 *
	 * os_delay(os_arg, microseconds)
	 * This function is supposed to delay or stall the processor for
	 * the passed in value number of microseconds.
	 */
	void *os_arg;
	int (*os_dmastream)(void *, unsigned char *, unsigned int);
	int (*os_dmaprep)(void *, unsigned char *, unsigned int);
	void (*os_delay)(void *, unsigned int);

	/* If the SD card last successfully reset is write protected, this
	 * member will be non-zero.
	 */
	unsigned int sd_wprot;

	/* If this card may have been already initialized by TS-SDBOOT, place
	 * the magic token it placed in the EP93xx SYSCON ScratchReg1 here
	 * to avoid re-initialization.
	 */
	unsigned int sdboot_token;

	/* CRC hint for subsequent sdwrite() call, used to optimize
	 * write throughput while using DMA by pre-calculating CRC's for
	 * next write
	 */
	unsigned char *sd_crchint;

	/* The block size of the memory device.  Normally 512, but can be 1024
	 * for larger cards
	 */
	unsigned int sd_blocksize;

	/* Password for auto-unlocking in sdreset()
	 */
	unsigned char *sd_pwd;

	/* If the SD card was password locked, this will be non-zero.
	 */
	unsigned int sd_locked;

	/* Whether or not writes can be parked.
	 */
	unsigned int sd_writeparking;

	/* The rest of these members are for private internal use and should
	 * not be of interest to client code.
	 */
	unsigned int sd_rcaarg;
	unsigned int sd_csd[17];
	unsigned int sd_cid[17];
	unsigned int sd_crcseq;
	unsigned short sd_crcs[4];
	unsigned int sd_crctmp[4];
	unsigned int sd_timeout;
	unsigned int parked_sector;
};

/* For sdreadv() / sdwritev() */
struct sdiov {
	unsigned char *sdiov_base;
	unsigned int sdiov_nsect;
};

int sdreset(struct sdcore *);
int sdread(struct sdcore *, unsigned int, unsigned char *, int);
int sdwrite(struct sdcore *, unsigned int, unsigned char *, int);
int sdreadv(struct sdcore *, unsigned int, struct sdiov *, int);
int sdwritev(struct sdcore *, unsigned int, struct sdiov *, int);
int sdsetwprot(struct sdcore *, unsigned int);
#define SDLOCK_UNLOCK	0
#define SDLOCK_SETPWD	1
#define SDLOCK_CLRPWD	2
#define SDLOCK_ERASE	8
int sdlockctl(struct sdcore *, unsigned int, unsigned char *, unsigned char *);

/*
 * Everything below here is secret!  This code shouldn't have to change
 * even for different OS.
 */

const static unsigned short crc16tbl[256] = {
	0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
	0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
	0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
	0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
	0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
	0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
	0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
	0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
	0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
	0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
	0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
	0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
	0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
	0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
	0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
	0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
	0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
	0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
	0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
	0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
	0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
	0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
	0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
	0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
	0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
};

const static unsigned char destagger[256] = {
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
	0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
	2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3,
};

static unsigned int crc7(unsigned int, const unsigned int *, unsigned int);
static int sdcmd(struct sdcore *, unsigned short, unsigned int, unsigned int *,
  unsigned char **);
static void mkcommand(unsigned int, unsigned int, unsigned int *);
static int stop(struct sdcore *);
static int do_read(struct sdcore *, unsigned int, struct sdiov *,
  unsigned int);
static int do_write(struct sdcore *, unsigned int, struct sdiov *,
  unsigned int);

#define POKE8(sd, x, y)	\
  *(volatile unsigned char *)((sd)->sd_regstart + (x)) = (y)

#define PEEK8(sd, x)	*(volatile unsigned char *)((sd)->sd_regstart + (x))

#define S_DUMMY_CLK	0
#define S_SEND_CMD	1
#define S_WAIT_RESP	2
#define S_RX_RESP	3
#define S_WAIT_BUSY	4
#define S_TX_WRITE	5
#define S_CRC_CHECK	6
#define S_OFF		7

#define TYPE_SHORTRESP	2
#define TYPE_LONGRESP	3
#define TYPE_BSYRESP	4
#define TYPE_NORESP	1
#define TYPE_RXDAT	0
#define TYPE_TXDAT	5
#define TYPE_ABORT	6

#define CMD(idx, type)	(0x40 | (idx) | ((type)<<8))

#define CMD_GO_IDLE_STATE		CMD(0, TYPE_NORESP)
#define CMD_ALL_SEND_CID		CMD(2, TYPE_LONGRESP)
#define CMD_SEND_RELATIVE_ADDR		CMD(3, TYPE_SHORTRESP)
#define CMD_SELECT_CARD			CMD(7, TYPE_BSYRESP)
#define CMD_DESELECT_CARD		CMD(7, TYPE_NORESP)
#define CMD_SEND_CSD			CMD(9, TYPE_LONGRESP)
#define CMD_PROGRAM_CSD			CMD(27, TYPE_TXDAT)
#define CMD_SET_BLOCKLEN		CMD(16, TYPE_SHORTRESP)
#define CMD_LOCK_UNLOCK			CMD(42, TYPE_TXDAT)
#define CMD_APP_CMD			CMD(55, TYPE_SHORTRESP)
#define CMD_READ_SINGLE_BLOCK		CMD(17, TYPE_RXDAT)
#define CMD_READ_MULTIPLE_BLOCK		CMD(18, TYPE_RXDAT)
#define CMD_STOP_TRANSMISSION		CMD(12, TYPE_ABORT)
#define CMD_SEND_STATUS			CMD(13, TYPE_SHORTRESP)
#define CMD_WRITE_BLOCK			CMD(24, TYPE_TXDAT)
#define CMD_WRITE_MULTIPLE_BLOCK	CMD(25, TYPE_TXDAT)

#define ACMD_SD_SEND_OP_COND		CMD(41, TYPE_SHORTRESP)
#define ACMD_SET_CLR_CARD_DETECT	CMD(42, TYPE_SHORTRESP)
#define ACMD_SET_BUS_WIDTH		CMD(6, TYPE_SHORTRESP)
#define ACMD_SET_WR_BLK_ERASE_COUNT	CMD(23, TYPE_SHORTRESP)
#define ACMD_SEND_NUM_WR_BLOCKS		CMD(22, TYPE_RXDAT)

/* Private bits for struct sdcore, sd_state member */
#define DATSSP_NOCRC		(1<<4)
#define DATSSP_4BIT		(1<<5)
#define PARKED_WRITE		(1<<6)

#define	NULL			((void *)0)


inline static unsigned short
crc16_acc(unsigned short crc, unsigned int b)
{
	return (crc << 8) ^ crc16tbl[(crc >> 8) ^ b];
}

static void sd_initcrc(struct sdcore *sd)
{
	int i;

	for (i = 0; i < 4; i++) {
		sd->sd_crctmp[i] = 0;
		sd->sd_crcs[i] = 0;
	}
	sd->sd_crcseq = 6;
}

static void sd_1bit_feedcrc(struct sdcore *sd, unsigned int dat)
{
	sd->sd_crcs[0] = crc16_acc(sd->sd_crcs[0], dat);
}

static void sd_4bit_feedcrc(struct sdcore *sd, unsigned int dat)
{
	unsigned int a = 0, b = 0, c = 0, d = 0;
	unsigned int shift = (sd->sd_crcseq & 0x7);

	a = sd->sd_crctmp[0];
	b = sd->sd_crctmp[1];
	c = sd->sd_crctmp[2];
	d = sd->sd_crctmp[3];

	a |= destagger[dat] << shift;
	dat >>= 1;
	b |= destagger[dat] << shift;
	dat >>= 1;
	c |= destagger[dat] << shift;
	dat >>= 1;
	d |= destagger[dat] << shift;

	if (shift == 0) {
		sd->sd_crcs[0] = crc16_acc(sd->sd_crcs[0], a);
		sd->sd_crcs[1] = crc16_acc(sd->sd_crcs[1], b);
		sd->sd_crcs[2] = crc16_acc(sd->sd_crcs[2], c);
		sd->sd_crcs[3] = crc16_acc(sd->sd_crcs[3], d);
		a = b = c = d = 0;
	}

	sd->sd_crcseq -= 2;
	sd->sd_crctmp[0] = a;
	sd->sd_crctmp[1] = b;
	sd->sd_crctmp[2] = c;
	sd->sd_crctmp[3] = d;
}

/* This should be called 8 times to get the full 8 bytes of CRC generated */
static unsigned int sd_4bit_getcrc(struct sdcore *sd)
{
	static const unsigned char restaggertbl[4] = { 0x0, 0x1, 0x10, 0x11 };
	static const unsigned char restaggertbl_lsl1[4] =
		{ 0x0, 0x2, 0x20, 0x22 };
	static const unsigned char restaggertbl_lsl2[4] =
		{ 0x0, 0x4, 0x40, 0x44 };
	static const unsigned char restaggertbl_lsl3[4] =
		{ 0x0, 0x8, 0x80, 0x88 };
	unsigned int ret;

	ret = restaggertbl[sd->sd_crcs[0] >> 14];
	sd->sd_crcs[0] <<= 2;
	ret |= restaggertbl_lsl1[sd->sd_crcs[1] >> 14];
	sd->sd_crcs[1] <<= 2;
	ret |= restaggertbl_lsl2[sd->sd_crcs[2] >> 14];
	sd->sd_crcs[2] <<= 2;
	ret |= restaggertbl_lsl3[sd->sd_crcs[3] >> 14];
	sd->sd_crcs[3] <<= 2;

	return ret;
}

/* This should be called 2 times to get the full 2 bytes of CRC generated */
static unsigned int sd_1bit_getcrc(struct sdcore *sd)
{
	unsigned int ret;

	ret = sd->sd_crcs[0] >> 8;
	sd->sd_crcs[0] = (sd->sd_crcs[0] & 0xff) << 8;
	return ret;
}

static inline void datssp_feedcrc(struct sdcore *sd, unsigned int dat)
{
	if (!(sd->sd_state & DATSSP_NOCRC)) {
		if (sd->sd_state & DATSSP_4BIT) sd_4bit_feedcrc(sd, dat);
		else sd_1bit_feedcrc(sd, dat);
	}
}

static inline unsigned int datssp_getcrc(struct sdcore *sd)
{
	unsigned int ret = 0;

	if (!(sd->sd_state & DATSSP_NOCRC)) {
		if (sd->sd_state & DATSSP_4BIT) ret = sd_4bit_getcrc(sd);
		else ret = sd_1bit_getcrc(sd);
	}
	return ret;
}

static inline unsigned int
crc7(unsigned int crc, const unsigned int *pc, unsigned int len)
{
	unsigned int i;
	unsigned char ibit;
	unsigned char c;

	for (i = 0; i < len; i++, pc++) {
		c = *pc;
		for (ibit = 0; ibit < 8; ibit++) {
			crc <<= 1;
			if ((c ^ crc) & 0x80) crc ^= 0x09;

			c <<= 1;
		}

		crc &= 0x7F;
	}

	return crc;
}

static inline void
mkcommand(unsigned int cmdidx, unsigned int arg, unsigned int *retcmd)
{
	retcmd[0] = cmdidx;
	retcmd[1] = arg >> 24;
	retcmd[2] = arg >> 16;
	retcmd[3] = arg >> 8;
	retcmd[4] = arg;
	retcmd[5] = (0x1 | (crc7(0, retcmd, 5) << 1));
}

static inline void reset_timeout(struct sdcore *sd) {
	sd->sd_timeout = 0;
}

static inline int timeout(struct sdcore *sd) {
	if (sd->sd_timeout > 1000000) return 1;
	else sd->sd_timeout++;
	return 0;
}

static unsigned int tend_ssp(struct sdcore *sd, unsigned int **cmdresp,
  unsigned char **dat) {
	unsigned int d;
	unsigned int s = PEEK8(sd, SDSTATE);

	if (s & 0x8) {
		if (sd->sd_state & SDCMD_RX) {
			d = PEEK8(sd, SDCMD);
			if (cmdresp) {
				**cmdresp = d;
				*cmdresp = *cmdresp + 1;
				reset_timeout(sd);
			}
		} else if (sd->sd_state & SDCMD_TX) {
			POKE8(sd, SDCMD, **cmdresp);
			*cmdresp = *cmdresp + 1;
			reset_timeout(sd);
		}
	}

	if (s & 0x10) {
		if (sd->sd_state & SDDAT_RX) {
			d = PEEK8(sd, SDDAT);
			if (dat) {
				**dat = d;
				*dat = *dat + 1;
				reset_timeout(sd);
			}
		} else if (sd->sd_state & SDDAT_TX) {
			reset_timeout(sd);
			if (dat) {
				d = **dat;
				*dat = *dat + 1;
				POKE8(sd, SDDAT, d);
				datssp_feedcrc(sd, d);
			} else {
				d = datssp_getcrc(sd);
				POKE8(sd, SDDAT, d);
			}
		}
	}

	return s;
}

static int
error(unsigned int *resp, unsigned short req)
{
	unsigned int crc, status;

	if ((req & 0x3f) != resp[0]) {
		return 1;
	}

	crc = (0x1 | (crc7(0, resp, 5) << 1));
	if (crc != resp[5]) {
		return 1;
	}

	status = resp[1] << 24;
	status |= resp[2] << 16;
	status |= resp[3] << 8;
	status |= resp[4];

	return status & 0xfdf90008;
}

static int
sdcmd(struct sdcore *sd, unsigned short req, unsigned int arg,
  unsigned int *resp, unsigned char **dat)
{
	unsigned int s, cmdresp[17];
	unsigned int resplen;
	unsigned int type = (req >> 8);
	unsigned int cmdidx = req;
	unsigned int *cmdptr = cmdresp;
	unsigned int *cmd = cmdresp;
	unsigned int *respptr;
	unsigned int ndat;

	// If no space for response provided by caller, use local buffer
	if (resp == NULL) resp = cmdresp;
	respptr = resp;

	// Before continuing, we must wait for the FSM to get to the
	// S_SEND_CMD state.  After a previous command, we may still be
	// in S_DUMMY_CLK or in case of an ABORT, we may be in the middle of
	// clocking a byte for TX or RX.
	s = PEEK8(sd, SDSTATE);
	while ((s & 0x7) != S_SEND_CMD) {
		if (timeout(sd)) break;
		s = PEEK8(sd, SDSTATE);
	}

	// We know we're in S_SEND_CMD, but we may need to change the
	// command type.  This won't cause a state change.
	if ((s & 0xe7) != (S_SEND_CMD | (type << 5)))
		POKE8(sd, SDSTATE, S_SEND_CMD | (type << 5));

	// Build command packet
	mkcommand(cmdidx, arg, cmdptr);

	// Next, we loop while tending the SSPs until we get our last
	// byte of command data out.  We may get a few bytes from the DAT
	// SSP if we are aborting a previous data transfer command.  If we do
	// those get placed in a buffer or thrown away based on the callers
	// "dat" parameter.
	sd->sd_state |= SDCMD_TX;
	while ((cmdptr - cmd) != 6) {
		if (timeout(sd)) break;
		s = tend_ssp(sd, &cmdptr, dat);
	}
	sd->sd_state &= ~SDCMD_TX;

	// If we got out of sync with the hardware, that would be bad.
	// The hardware should still be in S_SEND_CMD for the last CMDSSP
	// byte.
	if ((s & 0x7) != S_SEND_CMD) {
		POKE8(sd, SDSTATE, S_OFF);
		return 1;
	}

	if (type == TYPE_NORESP) goto done;
	else if (type == TYPE_LONGRESP) resplen = 17;
	else resplen = 6;

	// Next state should be S_WAIT_RESP or S_RX_RESP.  We may get
	// more bytes from the DATSSP while shifting out our last bits of cmd
	while (((s & 0x7) != S_WAIT_RESP) && ((s & 0x7) != S_RX_RESP)) {
		if (timeout(sd)) break;
		s = tend_ssp(sd, NULL, dat);
	}

	// Once we're in S_WAIT_RESP or S_RX_RESP though, the DATSSP is only
	// active for 2 more clocks at the beginning of the S_WAIT_RESP state.
	// This is enough for one more byte in 4-bit mode, though we may have
	// 2 bytes already in our DATSSP.
	if (sd->sd_state & (SDDAT_RX|SDDAT_TX)) {
		do {
			if (timeout(sd)) break;
			s = tend_ssp(sd, NULL, dat);
		} while (!(s & 0x18));

		// We've now read/wrote one more byte to the DATSSP
		// which should allow our FSM to advance to the RX_RESP state.
		// If we pick up more than 2 more DATSSP bytes, something is
		// wrong.
		ndat = 0;
		while ((s & 0x7) != S_RX_RESP) {
			if (timeout(sd) || ndat > 2) break;
			s = tend_ssp(sd, NULL, dat);
			if (s & 0x10) ndat++;
		}

		if (ndat > 2) {
			POKE8(sd, SDSTATE, S_OFF);
			return 1;
		}
	}

	// We're now done with whatever business we had remaining with the
	// previous command's DATSSP transfer since we've either just got our
	// first byte of response or our last byte of data
	sd->sd_state &= ~(SDDAT_RX|SDDAT_TX);
	if (type == TYPE_RXDAT) sd->sd_state |= SDDAT_RX;

	// Next we receive the response.  If this is TYPE_RXDAT command,
	// or an abortion of a previous TYPE_RXDAT command, we may get a
	// few bytes from the DAT SSP also.
	sd->sd_state |= SDCMD_RX;
	while ((respptr - resp) != resplen) {
		if (timeout(sd)) break;
		s = tend_ssp(sd, &respptr, dat);
		if ((s & 0x10) && (resp == respptr)) {
			POKE8(sd, SDSTATE, S_OFF);
			sd->sd_state &= ~(SDCMD_RX|SDDAT_RX);
			return 1;
		}
	}
	sd->sd_state &= ~SDCMD_RX;

	if (type == TYPE_ABORT)
		sd->sd_state &= ~(SDDAT_RX|SDDAT_TX);

	if (type == TYPE_TXDAT) sd->sd_state |= SDDAT_TX;

done:
	if (timeout(sd)) return 1;
	else return 0;

}

static int datssp_stream(struct sdcore *sd, unsigned char **dat,
  unsigned int buflen)
{
	unsigned int s, t, byte = 0;
	unsigned char *d;

	if (((sd->sd_state & SDDAT_RX) && sd->os_dmastream) /* ||
	  ((sd->sd_state & SDDAT_TX) && sd->os_dmastream && dat) */ ) {
		unsigned char *d = dat ? *dat : NULL;
		int ret = sd->os_dmastream(sd->os_arg, d, buflen);
		if (!ret && d) *dat += buflen;
		return ret;
	}

	while (buflen) {
		if (timeout(sd)) return 1;
		s = tend_ssp(sd, NULL, dat);
		if (s & 0x10) {
			buflen--;
			if (byte++ > 7) {
				if (sd->sd_state & SDDAT_RX)
				  goto fastrx;
				else goto fasttx;
			}
		}
	}

	// Now we can go faster (PIO)
fastrx:
	if (dat) {
		d = *dat;
		while (buflen) {
			s = PEEK8(sd, SDDAT);
			*d = s;
			buflen--;
			d++;
		}
		*dat = d;
	} else {
		while (buflen--) PEEK8(sd, SDDAT);
	}
	return 0;

fasttx:
	if (dat) {
		d = *dat;
		while (buflen) {
			t = *d;
			POKE8(sd, SDDAT, t);
			buflen--;
			d++;
			datssp_feedcrc(sd, t);
		}
		*dat = d;
	} else {
		while (buflen--) POKE8(sd, SDDAT, datssp_getcrc(sd));
	}
	return 0;
}

static int stop(struct sdcore *sd)
{
	int ret;
	unsigned int resp[6];

	if (sd->parked_sector) {
		if (sd->sd_state & SDDAT_TX) {
			/* wait to get out of S_WAIT_BUSY */
			while ((PEEK8(sd, SDSTATE) & 0x7) != S_TX_WRITE)
			  if (timeout(sd)) break;

			/* abort parked write */
			POKE8(sd, SDSTATE, S_SEND_CMD | (TYPE_ABORT << 5));
			sd->sd_state &= ~SDDAT_TX;
			sd->sd_state |= SDDAT_RX;
			ret = sdcmd(sd, CMD_STOP_TRANSMISSION, 0, resp, NULL);
			sd->sd_state &= ~SDDAT_RX;
			POKE8(sd, SDSTATE, S_WAIT_BUSY | (TYPE_BSYRESP << 5));
		} else {
			/* abort parked read */
			POKE8(sd, SDSTATE, S_SEND_CMD | (TYPE_ABORT << 5));
			ret = sdcmd(sd, CMD_STOP_TRANSMISSION, 0, resp, NULL);
		}
		sd->parked_sector = 0;
		if (ret || error(resp, CMD_STOP_TRANSMISSION) || timeout(sd)) return 1;
	}
	return 0;
}

static int do_read(struct sdcore *sd, unsigned int sector, struct sdiov *iov,
  unsigned int iovcnt)
{
	unsigned int resp[6], ret, n;
	unsigned char *datptr, *dat;

	if (iovcnt == 0) return 0;

	n = iov->sdiov_nsect;
	datptr = dat = iov->sdiov_base;

	if (sd->parked_sector) {
		if (!(sd->sd_state & SDDAT_TX) && sd->parked_sector == sector)
		  goto receive;

		ret = stop(sd);
		if (ret) return ret;
	}

	ret = sdcmd(sd, CMD_READ_MULTIPLE_BLOCK, sector * 512, resp, &datptr);
	if (ret || error(resp, CMD_READ_MULTIPLE_BLOCK)) return 1;

receive:
	if (sd->os_dmaprep && sd->os_dmastream)
	  sd->os_dmaprep(sd->os_arg, datptr, n * 512 - (datptr - dat));

	datssp_stream(sd, &datptr, 512 - (datptr - dat));
	datssp_stream(sd, NULL, 6);

	sd->parked_sector = sector + n;
	n--;

nextiov:
	while (n--) {
		POKE8(sd, SDSTATE, S_WAIT_RESP | (TYPE_RXDAT << 5));
		datssp_stream(sd, NULL, 2); // last part of prev CRC
		datssp_stream(sd, &datptr, 512);
		datssp_stream(sd, NULL, 6); // first part of CRC
	}

	if (--iovcnt) {
		++iov;
		n = iov->sdiov_nsect;
		datptr = iov->sdiov_base;
		sd->parked_sector += n;
		if (sd->os_dmaprep && sd->os_dmastream)
		  sd->os_dmaprep(sd->os_arg, datptr, n * 512);
		goto nextiov;
	}

	POKE8(sd, SDSTATE, S_WAIT_RESP | (TYPE_RXDAT << 5));
	datssp_stream(sd, NULL, 2); // last part of prev CRC
	return 0;
}

static int do_write(struct sdcore *sd, unsigned int sector, struct sdiov *iov,
  unsigned int iovcnt)
{
	unsigned char *datptr, *crcptr, **crcptrptr;
	unsigned int resp[6], ret, n;

	if (sd->sd_wprot) return 1;

	if (iovcnt == 0) return 0;

	if (0 /* sd->sd_crchint */) {
		// CRC is pre-calculated so don't recalculate
		crcptr = sd->sd_crchint;
		crcptrptr = &crcptr;
		sd->sd_state |= DATSSP_NOCRC;
		sd->sd_crchint = NULL;
	} else {
		crcptrptr = NULL;
		sd->sd_state &= ~DATSSP_NOCRC;
	}

	if (sd->parked_sector) {
		if ((sd->sd_state & SDDAT_TX) && sd->parked_sector == sector)
		  goto transmit;

		ret = stop(sd);
		if (ret) return ret;
	}

	if (sd->sd_erasehint) {
		sdcmd(sd, CMD_APP_CMD, sd->sd_rcaarg, NULL, NULL);
		sdcmd(sd, ACMD_SET_WR_BLK_ERASE_COUNT, sd->sd_erasehint,
		  NULL, NULL);
		sd->sd_erasehint = 0;
	}

	ret = sdcmd(sd, CMD_WRITE_MULTIPLE_BLOCK, sector * 512, resp, NULL);
	if (ret || error(resp, CMD_WRITE_MULTIPLE_BLOCK)) {
		return 1;
	}
	sd->parked_sector = sector;

transmit:
	while (iovcnt--) {
		datptr = iov->sdiov_base;
		n = iov->sdiov_nsect;
		sd->parked_sector += n;
		while (n--) {
			datssp_stream(sd, &datptr, 512);
			datssp_stream(sd, crcptrptr, 8); // CRC bytes
			POKE8(sd, SDSTATE, S_CRC_CHECK | (TYPE_TXDAT << 5));
		}
		iov++;
	}

	if (!sd->sd_writeparking) {
		ret = stop(sd);
		if (ret) return ret;
	}

	return 0;
}

static
unsigned int sdsize(struct sdcore *sd)
{
	unsigned int csize, csize_mult, rd_bl_len;

        rd_bl_len = 1 << ((sd->sd_csd[6] & 0xf) - 9);
	csize = ((sd->sd_csd[7] & 0x03) << 10) | ((sd->sd_csd[8] << 2) |
	  ((sd->sd_csd[9] & 0xc0) >> 6));
	csize_mult = ((sd->sd_csd[10] & 0x03) << 1) |
	  ((sd->sd_csd[11] & 0x80) >> 7);
	return (csize + 1) * (1 << (csize_mult + 2)) * rd_bl_len;
}

static
int sdfastinit(struct sdcore *sd)
{
	POKE8(sd, SDCTRL, 0x40);
	sd->sd_state = DATSSP_4BIT;

	sd->sd_rcaarg = ~sd->sdboot_token;
	sdcmd(sd, CMD_DESELECT_CARD, ~sd->sd_rcaarg, NULL, NULL);
	sdcmd(sd, CMD_SEND_CSD, sd->sd_rcaarg, sd->sd_csd, NULL);
	sdcmd(sd, CMD_SELECT_CARD, sd->sd_rcaarg, NULL, NULL);

	if (sd->os_dmastream) POKE8(sd, SDCTRL, 0x42);
	if ((PEEK8(sd, SDCTRL) & 0x80) || (sd->sd_csd[15] & 0x30))
	  sd->sd_wprot = 1;
	sd->sd_blocksize = 1 << ((sd->sd_csd[6] & 0xf));
	if (timeout(sd)) return 0;
	else return sdsize(sd);
}

int sdreset(struct sdcore *sd)
{
	unsigned int rca, s;
	unsigned int resp[17];

	// check for no SD card present
	if (PEEK8(sd, SDCTRL) & 0x8) return 0;

	reset_timeout(sd);
	sd_initcrc(sd);
	sd->parked_sector = 0;
	sd->sd_wprot = 0;
	sd->sd_blocksize = 0;

	if (sd->sdboot_token) {
		int ret = sdfastinit(sd);
		sd->sdboot_token = 0;
		if (ret) return ret;
	}

	// set controller for 1-bit mode, slow clock
	POKE8(sd, SDCTRL, 0x20);

	POKE8(sd, SDSTATE, S_DUMMY_CLK);
	sd->sd_state = SDCMD_RX|SDDAT_RX;
	s = PEEK8(sd, SDSTATE);
	while ((s & 0x7) != S_SEND_CMD) {
		// If we timeout here, it would be VERY BAD as we have no
		// further recourse to set things right if we can't turn
		// the SD off.
		if (timeout(sd)) return 0;
		sd->os_delay(sd->os_arg, 10000);
		sd->sd_timeout += 10000;

		// We won't be able to change state until both SSPs are empty
		s = tend_ssp(sd, NULL, NULL);
	}
	POKE8(sd, SDSTATE, S_OFF);
	sd->sd_state = 0;

	sd->os_delay(sd->os_arg, 50000);

	POKE8(sd, SDSTATE, S_DUMMY_CLK);
	sd->os_delay(sd->os_arg, 100000);
	if ((PEEK8(sd, SDSTATE) & 0x7) == S_OFF) {
		// No card present
		return 0;
	}

	POKE8(sd, SDSTATE, S_WAIT_RESP);
	// clock will freerun waiting for a response that will never come
	sd->os_delay(sd->os_arg, 50000);

	POKE8(sd, SDSTATE, S_DUMMY_CLK);

	do {
		sd->os_delay(sd->os_arg, 50000);
		sd->sd_timeout += 50000;
		sdcmd(sd, CMD_APP_CMD, 0, NULL, NULL);
		sdcmd(sd, ACMD_SD_SEND_OP_COND, 0x00ff0000, resp, NULL);
		if (timeout(sd)) break;
	} while (((resp[1] & 0x80) == 0x0));

	sdcmd(sd, CMD_ALL_SEND_CID, 0, sd->sd_cid, NULL);
	sdcmd(sd, CMD_SEND_RELATIVE_ADDR, 0, resp, NULL);
	rca = resp[1] << 8 | resp[2];
	sd->sd_rcaarg = (rca & 0xff00) << 16 | (rca & 0xff) << 16;

	sdcmd(sd, CMD_SEND_CSD, sd->sd_rcaarg, sd->sd_csd, NULL);
	sdcmd(sd, CMD_SELECT_CARD, sd->sd_rcaarg, resp, NULL);

	if ((resp[1] & 0x2)) {
		unsigned int ret = 1;
		sd->sd_locked = 1;
		if (sd->sd_pwd)
	 	  ret = sdlockctl(sd, SDLOCK_UNLOCK, sd->sd_pwd, NULL);
		if (ret != 0) return 0;
	} else sd->sd_locked = 0;

	sdcmd(sd, CMD_APP_CMD, sd->sd_rcaarg, NULL, NULL);
	sdcmd(sd, ACMD_SET_CLR_CARD_DETECT, 0, NULL, NULL);
	sdcmd(sd, CMD_SET_BLOCKLEN, 512, NULL, NULL);
	sdcmd(sd, CMD_APP_CMD, sd->sd_rcaarg, NULL, NULL);
	sdcmd(sd, ACMD_SET_BUS_WIDTH, 2, resp, NULL);

	// set controller for 4-bit mode, fast clock
	POKE8(sd, SDCTRL, (0x40 | (sd->os_dmastream ? 0x2 : 0x0)));
	sd->sd_state |= DATSSP_4BIT;

	if ((PEEK8(sd, SDCTRL) & 0x80) || (sd->sd_csd[15] & 0x30))
		sd->sd_wprot = 1;
	sd->sd_blocksize = 1 << ((sd->sd_csd[6] & 0xf));
	if (timeout(sd) || error(resp, ACMD_SET_BUS_WIDTH)) return 0;
	else return sdsize(sd);
}

int sdread(struct sdcore *sd, unsigned int sector, unsigned char *dat,
  int nsectors)
{
	struct sdiov iov;
	int ret;

	iov.sdiov_base = dat;
	iov.sdiov_nsect = nsectors;
	ret = do_read(sd, sector, &iov, 1);
	return ret;
}

int sdwrite(struct sdcore *sd, unsigned int sector, unsigned char *dat,
  int nsectors)
{
	struct sdiov iov;
	unsigned int ret;

	iov.sdiov_base = dat;
	iov.sdiov_nsect = nsectors;
	ret = do_write(sd, sector, &iov, 1);
	return ret;

}

int sdreadv(struct sdcore *sd, unsigned int sector, struct sdiov *iov,
  int niov)
{
	return do_read(sd, sector, iov, niov);
}

int sdwritev(struct sdcore *sd, unsigned int sector, struct sdiov *iov,
  int niov)
{
	return do_write(sd, sector, iov, niov);
}

int sdsetwprot(struct sdcore *sd, unsigned int perm)
{
	int i, ret;
	unsigned int csd[16], resp[6];
	unsigned char csdchars[16];
	unsigned char *csdptr = csdchars;

	if (stop(sd)) return 1;

	perm = perm ? 0x1 : 0x0;
	for (i = 0; i < 16; i++) csd[i] = sd->sd_csd[i + 1];
	csd[14] &= ~(0x3 << 4);
	csd[14] |= (perm << 4);
	csd[15] = 0x1 | crc7(0, csd, 15) << 1;
	for (i = 0; i < 16; i++) csdchars[i] = csd[i];

	ret = sdcmd(sd, CMD_PROGRAM_CSD, 0, resp, NULL);
	if (ret || error(resp, CMD_PROGRAM_CSD)) return 1;
	datssp_stream(sd, &csdptr, 16);
	datssp_stream(sd, NULL, 8);
	POKE8(sd, SDSTATE, S_CRC_CHECK | (TYPE_BSYRESP << 5));
	sd->sd_state &= ~SDDAT_TX;

	ret = sdcmd(sd, CMD_SEND_CSD, sd->sd_rcaarg, sd->sd_csd, NULL);
	if (ret || error(resp, CMD_SEND_CSD) ||
	  sd->sd_csd[15] != csd[14]) return 1;

	sd->sd_wprot = 1;
	return 0;
}

int sdlockctl(struct sdcore *sd, unsigned int cmd, unsigned char *pwd,
  unsigned char *sdbootdat)
{
	unsigned char pwddat[18];
	unsigned char *pwdptr = pwddat;
	unsigned int resp[6];
	int ret, i, len;
	int ccc = (sd->sd_csd[5] << 4) | (sd->sd_csd[6] >> 4);

	if (!(ccc & 0x80)) return 1; // Class 7 is lock-unlock commands

	if (pwd == NULL && cmd != SDLOCK_ERASE) return 1;

	if (stop(sd)) return 1;

	if (sd->sd_state & DATSSP_4BIT) {
		int oldctrl = PEEK8(sd, SDCTRL);
		int ret;

		sdcmd(sd, CMD_APP_CMD, sd->sd_rcaarg, NULL, NULL);
		sdcmd(sd, ACMD_SET_BUS_WIDTH, 0, NULL, NULL);
		POKE8(sd, SDCTRL, 0x20);
		sd->sd_state &= ~DATSSP_4BIT;
		ret = sdlockctl(sd, cmd, pwd, sdbootdat);
		sdcmd(sd, CMD_APP_CMD, sd->sd_rcaarg, NULL, NULL);
		sdcmd(sd, ACMD_SET_BUS_WIDTH, 2, NULL, NULL);
		sd->sd_state |= DATSSP_4BIT;
		POKE8(sd, SDCTRL, oldctrl);
		return ret;
	}

	pwddat[0] = cmd;
	if (cmd != SDLOCK_ERASE) {
		pwddat[1] = 16; // length
		for (i = 0; i < 16; i++) {
			pwddat[2 + i] = pwd[i];
		}
	}

	if (cmd == SDLOCK_ERASE) len = 1; else len = 18;
	ret = sdcmd(sd, CMD_SET_BLOCKLEN, len, resp, NULL);
	if (ret || error(resp, CMD_SET_BLOCKLEN)) return 1;
	ret = sdcmd(sd, CMD_LOCK_UNLOCK, 0, resp, NULL);
	if (ret || error(resp, CMD_LOCK_UNLOCK)) return 1;

	while ((pwdptr - pwddat) != len) {
		if (timeout(sd)) return 1;
		tend_ssp(sd, NULL, &pwdptr);
	}

	if (sd->sd_state & DATSSP_4BIT) datssp_stream(sd, NULL, 8);
	else datssp_stream(sd, NULL, 2);

	POKE8(sd, SDSTATE, S_CRC_CHECK | (TYPE_BSYRESP << 5));
	sd->sd_state &= ~SDDAT_TX;
	ret = sdcmd(sd, CMD_SET_BLOCKLEN, 512, resp, NULL);
	if (ret || error(resp, CMD_SET_BLOCKLEN)) return 1;
	ret = sdcmd(sd, CMD_SEND_STATUS, sd->sd_rcaarg, resp, NULL);
	if (ret || error(resp, CMD_SEND_STATUS)) return 1;

	if ((cmd == SDLOCK_ERASE || cmd == SDLOCK_UNLOCK ||
	  cmd == SDLOCK_CLRPWD) && (resp[1] & 0x2)) {
		return 1;
	}

	if (sdbootdat) {
		sdbootdat[0] = SDLOCK_UNLOCK;
		for (i = 1; i < 18; i++) {
			sdbootdat[i] = pwddat[i];
			sd_1bit_feedcrc(sd, pwddat[i]);
		}
		sdbootdat[18] = sd_1bit_getcrc(sd);
		sdbootdat[19] = sd_1bit_getcrc(sd);
	}

	return 0;
}
